#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include "server.h"

Server::Server(const char* host, uint16_t port) {
	const SSL_METHOD *method;
	
	method = SSLv2_server_method();
	
	ctx = SSL_CTX_new(method);
	if(!ctx) {
		fprintf(stderr, "error creating ctx\n");
		ERR_print_errors_fp(stderr);
		throw 0;
	}
	
	if(SSL_CTX_use_certificate_file(ctx, FILE_CERT, SSL_FILETYPE_PEM)<0) {
		fprintf(stderr, "error loading cert file\n");
		ERR_print_errors_fp(stderr);
		throw 0;
	}
	if(SSL_CTX_use_PrivateKey_file(ctx, FILE_KEY, SSL_FILETYPE_PEM)<0) {
		fprintf(stderr, "error loading private key file\n");
		ERR_print_errors_fp(stderr);
		throw 0;
	}
	
	if(!SSL_CTX_check_private_key(ctx)) {
		fprintf(stderr, "Key & certificate don't match\n");
		throw 0;
	}
	
	
	struct sockaddr_in sa;
	
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if(sock==-1) {
		fprintf(stderr, "Can't create socket\n");
		throw 0;
	}
	
	int opt = 1;
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt))==-1) {
		fprintf(stderr, "Can't set SO_REUSEADDR\n");
		throw 0;
	}
	
	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr.s_addr = inet_addr(host);
	
	if(bind(sock, (struct sockaddr*)&sa, sizeof(sa))==-1) {
		perror("bind");
		throw 0;
	}
	
	if(listen(sock, 10)==-1) {
		perror("listen");
		throw 0;
	}
	
	for(unsigned int i=0;i<MAX_CONN;i++) conns[i] = NULL;
}

void Server::loop() {
	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = 500;
	
	fd_set fdset;
	FD_ZERO(&fdset);
	FD_SET(1, &fdset);
	FD_SET(sock, &fdset);
	
	int max = sock;
	for(unsigned int i=0;i<MAX_CONN;i++) {
		if(conns[i]) {
			int h = conns[i]->h;
			if(h>max) max = h;
			FD_SET(h, &fdset);
		}
	}
	
	int res = select(max+1, &fdset, NULL, NULL, (struct timeval *)&tv);
	if(res<0) {
		if(errno==EINTR) return;
		perror("select");
		throw 0;
	}
	if(res==0) return;
	
	if(FD_ISSET(1, &fdset)) throw 0;
	if(FD_ISSET(sock, &fdset)) {
		try {
			struct sockaddr_in addr;
			socklen_t len = sizeof(addr);
			int client = accept(sock, (struct sockaddr *)&addr, &len);
			if(client==-1) {
				perror("accept");
				throw 0;
			}
			
			int k = -1;
			for(unsigned int i=0;i<MAX_CONN;i++) {
				if(conns[i]==NULL) {
					k = i;
					break;
				}
			}
			if(k==-1) {
				fprintf(stderr, "No place for new connection\n");
				shutdown(client, SHUT_RDWR);
				::close(client);
			} else {
				ClientConnection* c = new ClientConnection(client, &addr, ctx, this);
				conns[k] = c;
			}
		} catch(...) {}
	}
	
	for(unsigned int i=0;i<MAX_CONN;i++) {
		if(conns[i]) {
			int h = conns[i]->h;
			if(FD_ISSET(h, &fdset)) {
				conns[i]->loop();
			}
		}
	}
}

void Server::close() {
	for(unsigned int i=0;i<MAX_CONN;i++) {
		if(conns[i]) {
			conns[i]->close(false);
		}
	}
	::close(sock);
}

void Server::broadcast(uint8_t* data, size_t size, ClientConnection* me) {
	for(unsigned int i=0;i<MAX_CONN;i++) {
		if(!conns[i]) continue;
		if(conns[i]==me) continue;
		conns[i]->write(data, size);
	}
}

void Server::delete_me(ClientConnection* c) {
	for(unsigned int i=0;i<MAX_CONN;i++) {
		if(!conns[i]) continue;
		if(conns[i]!=c) continue;
		delete conns[i];
		conns[i] = NULL;
		printf("deleted\n");
	}
}

